#ifndef lint
static char SCCSid[] = "%W% %G%";
#endif

#include "mpid.h"

/* 
  This file contains the routines to handle receiving a message

  Because we don't know the length of messages (at least the long ones with
  tag MPID_PT2PT2_TAG(src) tags), we never post a receive.  Rather, we have
  a MPID_ch_check_incoming routine that looks for headers.  Note that
  messages sent from a source to destination with the MPID_PT2PT2_TAG(src)
  are ordered (we assume that the message-passing system preserves order).
 */


/*
   Post a receive.

   Since the Chameleon implementation lets the underlying message transport
   layer handle this, there isn't much to do.  We DO handle the case of
   posting a BLOCKING receive by doing BLOCKING waits on incoming messages.

   Otherwise, we simply try to handle any receives that are ready for
   processing.
 */
int MPID_CH_post_recv( mpid_recv_handle, dmpi_recv_handle ) 
MPID_RHANDLE *mpid_recv_handle;
MPIR_RHANDLE *dmpi_recv_handle;
{
/* First, see if the message is already here... */
if (mpid_recv_handle->temp) {
    DMPI_put_from_contig( dmpi_recv_handle, (void *)(mpid_recv_handle->temp),
   	                  mpid_recv_handle->len );
    FREE( mpid_recv_handle->temp );
    mpid_recv_handle->temp = 0;      /* In case of a cancel */
    mpid_recv_handle->done = YES;
    DMPI_mark_recv_completed(dmpi_recv_handle);
    }

else if (!mpid_recv_handle->is_non_blocking) {
    while (!mpid_recv_handle->done) {
	(void)MPID_CH_check_incoming( MPID_BLOCKING );
	}
    }

else
    while (MPID_CH_check_incoming( MPID_NOTBLOCKING ) == -1) ;

return MPI_OK;
}

/* Check for incoming messages.
    Input Parameter:
.   is_blocking - true if this routine should block until a message is
    available

    Returns -1 if nonblocking and no messages pending
 */    
int MPID_CH_check_incoming( is_blocking )
MPID_BLOCKING_TYPE is_blocking;
{
MPID_PACKET  pkt;	
int          from;
char         *address;
MPIR_RHANDLE *dmpi_recv_handle;
int          actual_len;
MPID_RHANDLE *mpid_recv_handle;
int          status;

/* If nonblocking and no headers available, exit */
if (is_blocking == MPID_NOTBLOCKING) {
    if (!PInprobe( MPID_PT2PT_TAG )) return -1;
    }

PIbrecv( MPID_PT2PT_TAG, &pkt, sizeof(MPID_PACKET), MSG_OTHER );
from = PIfrom();
/* We should check the size here for internal errors .... */

/* Is the message expected or not? */
DMPI_msg_arrived( from, pkt.tag, pkt.context_id, 
                  &dmpi_recv_handle, &status );
DMPI_mpid_recv_handle_from_rhandle( dmpi_recv_handle, mpid_recv_handle );
if (status) {
    DMPI_put_from_contig( dmpi_recv_handle, (void *)(pkt.buffer),
   	                  MPID_MIN(pkt.len,MPID_PACKET_SIZE) );
    /* Handle long messages... */
    if (pkt.len > MPID_PACKET_SIZE) {
        /* Question: who is responsible for this allocation? */
	DMPI_put_into_contig( dmpi_recv_handle, (void **)(&address), 
			      pkt.len - MPID_PACKET_SIZE, &actual_len );
        PIbrecv( MPID_PT2PT2_TAG(from), address, pkt.len-MPID_PACKET_SIZE,
   	         MSG_OTHER );
        }
    DMPI_mark_recv_completed(dmpi_recv_handle);
    mpid_recv_handle->done = YES;
    }
else {
    /* We buffer the message on the receiver.   There are, of course, other
       strategies */
    address = (char *)MALLOC( pkt.len ); 
    if (!address) return MPI_EXAUSTED;
    memcpy( address, pkt.buffer, MPID_MIN(pkt.len,MPID_PACKET_SIZE) );
    if (pkt.len > MPID_PACKET_SIZE) {
	/* Message has "extension" part */
        PIbrecv( MPID_PT2PT2_TAG(from), address + MPID_PACKET_SIZE, 
		 pkt.len-MPID_PACKET_SIZE, MSG_OTHER );
	}
    mpid_recv_handle->temp = address;
    }
return MPI_OK;    
}


/*
    This routine completes a particular receive.  It does this by processing
    incoming messages until the indicated message is received
 */
int MPID_CH_complete_recv( mpid_recv_handle, dmpi_recv_handle ) 
MPID_RHANDLE *mpid_recv_handle;
MPIR_RHANDLE *dmpi_recv_handle;
{
while (!mpid_recv_handle->done) {
    (void)MPID_CH_check_incoming( MPID_BLOCKING );
    }
return MPI_OK;
}
